UNIT FFT;

INTERFACE

CONST
    max_n_samples    = 1024;
    max_n_samples_m1 = 1023;
    max_n_spec       =  512;

TYPE
    complex = RECORD re, im: SINGLE END;
    fourier_signal = ARRAY [0 .. max_n_samples_m1] OF SINGLE;
    fourier_spec   = ARRAY [0 .. max_n_spec]       OF complex;
    fourier_mag    = ARRAY [0 .. max_n_spec]       OF SINGLE;
    fourier_record = RECORD CASE INTEGER OF
        1: (s: fourier_signal);
        2: (a: fourier_spec);
    END;


FUNCTION TwoTo(val: INTEGER): INTEGER;

PROCEDURE ComplexMult(VAR a0, a: complex);

PROCEDURE ComplexSqrt(VAR a0: complex);

PROCEDURE ComplexRecip(VAR a0: complex);

PROCEDURE ClearFourierSpec(VAR a: fourier_spec; m: INTEGER);

PROCEDURE ClearFourierSignal(VAR s: fourier_signal; m: INTEGER);

PROCEDURE ComplexFourierTransform(VAR a: fourier_spec; m: INTEGER);

PROCEDURE ComplexInverseTransform(VAR a: fourier_spec; m: INTEGER);

PROCEDURE RealFourierTransform(VAR frec: fourier_record; m: INTEGER);

PROCEDURE RealInverseTransform(VAR frec: fourier_record; m: INTEGER);


IMPLEMENTATION

TYPE
   {The use of double precision helps the generation of the butterfly
    twiddle factors by a series of complex multiplies from losing accuracy.}
    long_complex = RECORD re, im: DOUBLE END;


PROCEDURE ComplexMult(VAR a0, a: complex);

    VAR temp: complex;

BEGIN
    temp := a0;
    a0.re := temp.re*a.re - temp.im*a.im;
    a0.im := temp.re*a.im + temp.im*a.re;
END;


PROCEDURE ComplexSqrt(VAR a0: complex);

    VAR phase, magsqrt: SINGLE;

BEGIN
    WITH a0 DO BEGIN
        IF re > 0 THEN
            phase := (arctan(im/re))/2.0
        ELSE IF re < 0 THEN BEGIN
            IF im >= 0 THEN
                phase := (arctan(im/re) + pi)/2.0
            ELSE
                phase := (arctan(im/re) - pi)/2.0
        END ELSE IF im > 0 THEN
            phase :=  pi/2
        ELSE IF im < 0 THEN
            phase := -pi/2
        ELSE
            phase := 0;
        magsqrt := Sqrt(Sqrt(re*re + im*im));
        re := magsqrt*cos(phase);
        im := magsqrt*sin(phase);
    END;
END;


PROCEDURE ComplexRecip(VAR a0: complex);

    VAR magsq: SINGLE;

BEGIN
    WITH a0 DO BEGIN
        magsq := re*re + im*im;
        re := re/magsq;
        im := -im/magsq;
    END;
END;


PROCEDURE LComplexMult(VAR a0, a: long_complex);
   {This procedure is used to generate the butterfly twiddle factors}
    VAR temp: long_complex;

BEGIN
    temp := a0;
    a0.re := temp.re*a.re - temp.im*a.im;
    a0.im := temp.re*a.im + temp.im*a.re;
END;


PROCEDURE Butterfly(VAR aj, ak: complex; wval: long_complex);

    VAR temp: complex;

BEGIN
    temp.re := ak.re*wval.re - ak.im*wval.im;
    temp.im := ak.im*wval.re + ak.re*wval.im;
    ak.re := aj.re - temp.re;
    ak.im := aj.im - temp.im;
    aj.re := aj.re + temp.re;
    aj.im := aj.im + temp.im;
END;


FUNCTION TwoTo(val: INTEGER): INTEGER;

    VAR i, result: INTEGER;

BEGIN
    result := 1;
    FOR i := 1 TO val DO result := result*2;
    TwoTo := result;
END;


PROCEDURE ClearFourierSpec(VAR a: fourier_spec; m: INTEGER);

    VAR i: INTEGER;

BEGIN
    FOR i := 0 TO TwoTo(m) DO BEGIN
        a[i].re := 0;
        a[i].im := 0;
    END;
END;


PROCEDURE ClearFourierSignal(VAR s: fourier_signal; m: INTEGER);

    VAR i: INTEGER;

BEGIN
    FOR i := 0 TO TwoTo(m)-1 DO s[i] := 0;
END;


PROCEDURE ComplexFourierTransform(VAR a: fourier_spec; m: INTEGER);

    VAR
        n, nv2, jj, i, j, k, l, lev2v2, lev2, le, wcount: INTEGER;
        temp: complex;
        wval, dwval: long_complex; {wval is the butterfly twiddle factor}
        dtheta: DOUBLE;

BEGIN
    n := TwoTo(m);

    j := 0;
    nv2 := n DIV 2;
    FOR i := 0 TO n-2 DO BEGIN
        IF i < j THEN BEGIN
            temp := a[i];
            a[i] := a[j];
            a[j] := temp;
        END;
        k := nv2;
        WHILE k <= j DO BEGIN
            j := j - k;
            k := k DIV 2;
        END;
        j := j + k;
    END;

    dtheta := 2.0*Pi;
    le := 1;
    lev2 := 0;

    FOR l := 1 TO m DO BEGIN
        lev2v2 := lev2;
        lev2 := le;
        le := le * 2;

        wval.re := 1.0;
        wval.im := 0.0;
        wcount := 0;
        dtheta := dtheta / 2.0;

        dwval.re := cos(dtheta);
        dwval.im := -sin(dtheta);

        jj := 0;

        FOR i := 0 TO lev2-1 DO BEGIN
            j := jj;
            WHILE j < n DO BEGIN
                k := j + lev2;
                Butterfly(a[j],a[k],wval);
                j := j + le;
            END;

            jj := jj + 1;

            wcount := wcount + 1;
            LComplexMult(wval,dwval);
        END;

    END;
END;


PROCEDURE ComplexInverseTransform(VAR a: fourier_spec; m: INTEGER);

    VAR
        i, n: INTEGER;
        fn: SINGLE;

BEGIN
    n := TwoTo(m);
    fn := n;
    FOR i := 0 TO n-1 DO BEGIN
        a[i].im := -a[i].im;
    END;
    ComplexFourierTransForm(a,m);
    FOR i := 0 TO n-1 DO BEGIN
        a[i].re :=  a[i].re/fn;
        a[i].im := -a[i].im/fn;
    END;
END;


PROCEDURE RealFourierTransform(VAR frec: fourier_record; m: INTEGER);

    VAR nv2, nv4, k: INTEGER;
        long_temp, wval, dwval: long_complex;
        xe, xo, wxo: complex;

BEGIN
    ComplexFourierTransform(frec.a,m-1);
    nv2 := TwoTo(m-1);
    nv4 := nv2 DIV 2;
    frec.a[nv2].im := 0;
    frec.a[nv2].re := frec.a[0].re - frec.a[0].im;
    frec.a[0].re   := frec.a[0].re + frec.a[0].im;
    frec.a[0].im := 0;

    wval.re := 1;
    wval.im := 0;
    dwval.re := Cos(Pi/nv2);
    dwval.im := -Sin(Pi/nv2);
    FOR k := 1 TO nv4 DO WITH frec DO BEGIN
        LComplexMult(wval,dwval);
        xe.re := 0.5*(a[k].re + a[nv2-k].re);
        xe.im := 0.5*(a[k].im - a[nv2-k].im);
        xo.re := 0.5*(a[k].im + a[nv2-k].im);
        xo.im := 0.5*(a[nv2-k].re - a[k].re);
        wxo.re := xo.re*wval.re - xo.im*wval.im;
        wxo.im := xo.im*wval.re + xo.re*wval.im;
        a[k].re     :=  xe.re + wxo.re;
        a[k].im     :=  xe.im + wxo.im;
        a[nv2-k].re :=  xe.re - wxo.re;
        a[nv2-k].im := -xe.im + wxo.im;
    END;
END;


PROCEDURE RealInverseTransform(VAR frec: fourier_record; m: INTEGER);

    VAR nv2, nv4, k: INTEGER;
        long_temp, wval, dwval: long_complex;
        xe, xo, wxo: complex;

BEGIN
    nv2 := TwoTo(m-1);
    nv4 := nv2 DIV 2;
    frec.a[0].re := 0.5*(frec.a[0].re + frec.a[nv2].re);
    frec.a[0].im := frec.a[0].re - frec.a[nv2].re;
    wval.re := 1;
    wval.im := 0;
    dwval.re :=  Cos(Pi/nv2);
    dwval.im := -Sin(Pi/nv2);
    FOR k := 1 TO nv4 DO WITH frec DO BEGIN
        LComplexMult(wval,dwval);
        xe.re  := 0.5*(a[k].re + a[nv2-k].re);
        xe.im  := 0.5*(a[k].im - a[nv2-k].im);
        wxo.re := 0.5*(a[k].re - a[nv2-k].re);
        wxo.im := 0.5*(a[k].im + a[nv2-k].im);
        xo.re := wxo.re*wval.re + wxo.im*wval.im;
        xo.im := wxo.im*wval.re - wxo.re*wval.im;
        a[k].re     := xe.re - xo.im;
        a[k].im     := xe.im + xo.re;
        a[nv2-k].re := xe.re + xo.im;
        a[nv2-k].im := xo.re - xe.im;
    END;

    ComplexInverseTransform(frec.a,m-1);
END;

END.
