UNIT LPC; {Copyright 1992 Paul Milenkovic ALL RIGHTS RESERVED}

INTERFACE USES DSPTypes, DSPFast, LinEq;

TYPE
    predictor     = DSPTypes.coef_v0;
        {The predictor coefficients are defined for variable a of type
         predictor with a[0] = 1.0 where the LPC inverse filter applied
         to a signal s[i] produces output

             e[i] := a[0]*s[i] + a[1]*s[i-1] + ... + a[n]*s[i-n]

         and the LPC forward filter applied to e[i] produces the output

             s[i] := e[i] - (a[1]*s[i-1] + ... + a[n]*s[i-n])

        }
    sos_list      = ARRAY [1 .. DSPTypes.mcoefv2] OF RECORD p, q: SINGLE END;


PROCEDURE CrossLPC(

    VAR s0, s: sig_buffer; VAR a: predictor;
    nsig, ncoef: INTEGER; snr: SINGLE; VAR ssqer: SINGLE
);

PROCEDURE CovarianceLPC(

    VAR s: sig_buffer; VAR a: predictor;
    nsig, ncoef: INTEGER; snr: SINGLE; VAR ssqer: SINGLE
);

PROCEDURE AutocorrLPC(

    VAR s: sig_buffer; VAR a: predictor;
    nsig, ncoef: INTEGER; snr: SINGLE; VAR ssqer: SINGLE
);

PROCEDURE CovMat(

    VAR s: sig_buffer; VAR phi: cov_matrix; nsig, ncoef: INTEGER
);

PROCEDURE Cholesky(

    VAR phi: cov_matrix; VAR a: predictor; ncoef: INTEGER;
    greq, snr: SINGLE; VAR ssqer: SINGLE
);

PROCEDURE LPCtoSOS(

    VAR a0: predictor; ncoef: INTEGER; VAR sos: sos_list; VAR nsos: INTEGER
);

PROCEDURE SOStoLPC(

    VAR sos: sos_list; nsos: INTEGER; VAR a: predictor; VAR ncoef: INTEGER
);

PROCEDURE SOStoFBW(

    VAR sos: sos_list; nsos: INTEGER; ms: DOUBLE;
    VAR fbw: sos_list; VAR nfbw: INTEGER
);

FUNCTION  LPCtoLog2Area(VAR a: predictor; ncoef: INTEGER): BOOLEAN;

PROCEDURE Log2AreatoLPC(VAR a: predictor; ncoef: INTEGER);


IMPLEMENTATION

PROCEDURE CrossLPC(

    VAR s0, s: sig_buffer; VAR a: predictor;
    nsig, ncoef: INTEGER; snr: SINGLE; VAR ssqer: SINGLE
);

    VAR i, j: INTEGER;
        g: Lineq.g_list;
        b: Lineq.g_col;
        alpha: SINGLE;

BEGIN
    FOR j := 1 TO ncoef DO BEGIN
{
        b[j] := 0;
        FOR i := 0 TO Pred(nsig) DO b[j] := b[j] + s[i-1]*s[i-j];
}
        b[Pred(j)] := DotProduct(s[-1],s[-j],nsig);
    END;
    b[0]    := b[0]*(1 + 1/snr);
    alpha   := Sqrt(b[0]);
    g[1][1] := alpha;
    g[2][1] := 0;
    g[3][1] := 0;
    g[4][1] := 0;
    alpha   := 1/alpha;
    g[1][0] := alpha;
    g[2][0] := 0;
    g[3][0] := alpha;
    g[4][0] := 0;
    FOR j := 2 TO ncoef DO BEGIN
        g[1][j] := b[j-1]*alpha;
        g[2][j] := s[-j];
        g[3][j] := g[1][j];
        g[4][j] := s[nsig-j];
    END;
    b[0] := DotProduct(s0[0],s0[0],nsig);
    FOR j := 1 TO ncoef DO
        b[j] := DotProduct(s0[0],s[-j],nsig);
    FastSolve(g,b,2,2,ncoef);
    a[0] := 1;
    FOR j := 1 TO ncoef DO
        a[j] := b[j-1];
    ssqer := b[ncoef];
END;


PROCEDURE CovarianceLPC(

    VAR s: sig_buffer; VAR a: predictor;
    nsig, ncoef: INTEGER; snr: SINGLE; VAR ssqer: SINGLE
);

    VAR i, j: INTEGER;
        g: Lineq.g_list;
        b: Lineq.g_col;
        alpha: SINGLE;

BEGIN
    FOR j := 0 TO ncoef DO BEGIN
{
        b[j] := 0;
        FOR i := 0 TO Pred(nsig) DO b[j] := b[j] + s[i]*s[i-j];
}
        b[j] := DotProduct(s[0],s[-j],nsig);
    END;
    FOR j := 0 TO ncoef DO BEGIN
        IF j = 0 THEN BEGIN
            b[0] := b[0]*(1 + 1/snr);
            alpha := Sqrt(b[0]);
            g[1][1] := alpha;
            g[3][1] := 0;
            alpha := 1/alpha;
            g[1][0] := alpha;
            g[2][0] := 0;
            g[3][0] := alpha;
            g[4][0] := 0;
        END ELSE BEGIN
            g[2][j] := s[-j];
            g[4][j] := s[nsig-j];
        END;
        IF j > 1 THEN BEGIN
            g[1][j] := b[j-1]*alpha;
            g[3][j] := g[1][j];
        END;
    END;
    FastSolve(g,b,2,2,ncoef);
    a[0] := 1;
    FOR j := 1 TO ncoef DO a[j] := b[j-1];
    ssqer := b[ncoef];
END;


PROCEDURE AutocorrLPC(

    VAR s: sig_buffer; VAR a: predictor;
    nsig, ncoef: INTEGER; snr: SINGLE; VAR ssqer: SINGLE
);

    VAR i, j: INTEGER;
        g: Lineq.g_list;
        b: Lineq.g_col;
        alpha, sum: SINGLE;

BEGIN
    FOR j := 0 TO ncoef DO BEGIN
{
        b[j] := 0;
        FOR i := j TO Pred(nsig) DO b[j] := b[j] + s[i]*s[i-j];
 }
        b[j] := DotProduct(s[j],s[0],nsig-j);
    END;
    alpha := Sqrt(b[0]*(1 + 1/snr));
    g[1][1] := alpha;
    g[2][1] := 0;
    alpha := 1/alpha;
    g[1][0] := alpha;
    g[2][0] := alpha;
    FOR j := 2 TO ncoef DO BEGIN
        g[1][j] := b[j-1]*alpha;
        g[2][j] := g[1][j];
    END;
    FastSolve(g,b,1,1,ncoef);
    a[0] := 1;
    FOR j := 1 TO ncoef DO a[j] := b[j-1];
    ssqer := b[ncoef];
END;


PROCEDURE CovMat(

    VAR s: sig_buffer; VAR phi: cov_matrix; nsig, ncoef: INTEGER
);

    {Computes the covariance matrix phi from the signal sample values
     s[k] for covariance method LPC analysis of order ncoef.  The least
     squares interval on s[k] starts at k=0 and ends at k=nsig-1.
     Be sure to specify values for s[-ncoef] through s[nsig-1].}

    VAR
        sum: DOUBLE;
        k, col, lag, delay, n1, n2: INTEGER;

BEGIN
    FOR col := 0 TO ncoef DO BEGIN
        lag := col;
        sum := 0;
        FOR k := 0 TO Pred(nsig) DO sum := sum + s[k]*s[k-lag];
        phi[0,col] := sum;
        phi[col,0] := sum;
        FOR delay := 1 TO ncoef-lag DO BEGIN
            n1 :=      - delay;
            n2 := nsig - delay;
            sum := sum + s[n1]*s[n1-lag] - s[n2]*s[n2-lag];
            phi[delay,delay+col] := sum;
            phi[delay+col,delay] := sum;
        END;
    END;
END;


PROCEDURE Cholesky(

    VAR phi: cov_matrix; VAR a: predictor; ncoef: INTEGER;
    greq, snr: SINGLE; VAR ssqer: SINGLE
);

    {Computes predictor coefficients a from covariance matrix phi up to
     order ncoef.  The variable greq is the maximum allowable prediction
     gain (set to 5000.0 for speech).  The prediction gain is the
     ratio in energy of signal s[n] to LPC residual e[n], and it is used
     to limit the LPC order to a value where the LPC coefficients are
     not corrupted by the numerical illconditioning of extremely high levels
     of prediction gain.  Where the order is limited, a[n] will be zero filled.

     This routine expects the phi matrix to be symmetric.  It uses the
     lower triangular part as a work area, but it restores phi to its
     original condition using the values stored in the upper triangular
     part.  In this sense, it is nondestructive of the phi matrix data.}

    VAR
        phi00, ep, sum, aii, gact: SINGLE;
        lcoef, i, j, k, ii: INTEGER;
        limits_ok : BOOLEAN;
        phijj, phii0: predictor;

BEGIN
    a[0]  := 1.0;
    FOR i := 1 TO ncoef DO a[i] := 0;
    phi00 := phi[0,0];
    FOR i := 0 TO ncoef DO phijj[i] := phi[i,i] + phi00/snr;
    phii0[0] := 0;
    FOR i := 1 TO ncoef DO phii0[i] := phi[i,0];

    lcoef := 0;
    gact  := 1.0;
    phi00 := phi00*(1 + 1/snr);
    ep    := phi00;

    limits_ok := true;

    FOR i := 1 TO ncoef DO IF limits_ok THEN BEGIN
        FOR j := 1 TO i-1 DO BEGIN
            FOR k := 1 TO j-1 DO
                phi[i,j] := phi[i,j] - phi[i,k]*phijj[k]*phi[j,k];
            phi[i,j] := phi[i,j]/phijj[j];
        END;
        FOR j := 1 TO i-1 DO BEGIN
            phijj[i] := phijj[i] - phi[i,j]*phi[i,j]*phijj[j];
            phii0[i] := phii0[i] - phi[i,j]*phii0[j];
        END;
        IF phijj[i] < 1.0e-20*phi00 THEN BEGIN
            limits_ok := false;
            gact := -1.0;
        END ELSE BEGIN
            ep := ep - phii0[i]*phii0[i]/phijj[i];
            IF ep > 1.0e-6*phi00 THEN gact := phi00/ep ELSE gact := greq;
            lcoef := i;
            IF gact >= greq THEN limits_ok := false;
        END;
    END;

    FOR i := 1 TO lcoef DO BEGIN
        ii := Succ(lcoef - i);
        aii := -phii0[ii]/phijj[ii];
        IF ii < lcoef THEN FOR j := ii+1 TO lcoef DO BEGIN
            aii := aii - phi[j,ii]*a[j];
        END;
        a[ii] := aii;
    END;

    FOR i := 1 TO ncoef DO FOR j := 0 TO i-1 DO phi[i,j] := phi[j,i];
    ssqer := ep;
END;


PROCEDURE LPCtoSOS(

    VAR a0: predictor; ncoef: INTEGER; VAR sos: sos_list; VAR nsos: INTEGER
);

    VAR iter, k, m, lsos, i: INTEGER;
        ck, ckm1, ckm2, bk, bkm1, bkm2: DOUBLE;
        bm, bmm1, cmm1, cmm2, cmm3, cmbvc, dq, dp: DOUBLE;
        a: predictor;
        thresh: DOUBLE;

BEGIN
    lsos   := 0;
    m      := ncoef;
    a      := a0;
    thresh := 1e-7;

    WHILE m > 2 DO BEGIN
        Inc(lsos);
        WITH sos[lsos] DO BEGIN
            IF nsos < lsos THEN BEGIN
                nsos := lsos;
                thresh := 1e-5;
                p := 0;
                q := 0;
            END;
            iter := 0;
            REPEAT
                bkm1 := 1;
                ckm1 := 1;
                bk   := a[1] - p;
                ck   := bk   - p;
                FOR k := 2 TO Pred(m) DO BEGIN
                    bkm2 := bkm1;
                    bkm1 := bk;
                    bk   := a[k] - p*bkm1 - q*bkm2;
                    ckm2 := ckm1;
                    ckm1 := ck;
                    ck   := bk   - p*ckm1 - q*ckm2;
                END;
                bmm1  := bk;
                bm    := a[m] - p*bk - q*bkm1;
                cmm3  := ckm2;
                cmm2  := ckm1;
                cmm1  := ck;
                cmbvc := (cmm1-bmm1)/cmm2;
                dq    := (bm-bmm1*cmbvc)/(cmm2-cmm3*cmbvc);
                dp    := (bmm1-cmm3*dq)/cmm2;
                p     := p + dp;
                q     := q + dq;
                Inc(iter);
            UNTIL ((Abs(dp) < thresh) AND (Abs(dq) < thresh)) OR (iter >= 1000);
            IF (iter >= 1000) AND ((Abs(dp) > 1e-3) OR (Abs(dq) > 1e-3)) THEN
                m := 0
            ELSE BEGIN
                m    := m - 2;
                bk   := 1;
                bkm1 := 0;
                FOR k := 1 TO m DO BEGIN
                    bkm2 := bkm1;
                    bkm1 := bk;
                    bk   := a[k] - p*bkm1 - q*bkm2;
                    a[k] := bk;
                END;
            END;
        END;
    END;
    IF m > 0 THEN BEGIN
        Inc(lsos);
        IF nsos < lsos THEN nsos := lsos;
        WITH sos[lsos] DO BEGIN
            p := a[1];
            IF m > 1 THEN q := a[2] ELSE q := 0;
        END;
    END;
END;


PROCEDURE SOStoLPC(

    VAR sos: sos_list; nsos: INTEGER; VAR a: predictor; VAR ncoef: INTEGER
);

    VAR i, j: INTEGER;

BEGIN
    a[0]  := 1.0;
    ncoef := 0;
    FOR i := 1 TO nsos DO BEGIN
        ncoef      := ncoef + 2;
        a[ncoef]   := 0;
        a[ncoef-1] := 0;
        FOR j := ncoef DOWNTO 1 DO WITH sos[i] DO BEGIN
            a[j] := a[j] + p*a[j-1];
            IF j > 1 THEN a[j] := a[j] + q*a[j-2];
        END;
   END;
END;


PROCEDURE SOStoFBW(

    VAR sos: sos_list; nsos: INTEGER; ms: DOUBLE;
    VAR fbw: sos_list; VAR nfbw: INTEGER
);

    VAR i: INTEGER;
        dscr, fq, bw: DOUBLE;

BEGIN
    nfbw := 0;
    FOR i := 1 TO nsos DO WITH sos[i] DO BEGIN
        dscr := p*p/4 - q;
        IF dscr < 0 THEN BEGIN
            IF p = 0 THEN
                fq := Pi/2
            ELSE
                fq := ArcTan(-2*Sqrt(-dscr)/p);
            IF fq < 0 THEN fq := Pi + fq;
            fq :=  fq/(2*Pi*ms);
            bw := -ln(q)/(2*Pi*ms);
            Inc(nfbw);
            fbw[nfbw].p := fq;
            fbw[nfbw].q := bw;
{
        END ELSE BEGIN
            dscr := Sqrt(dscr);
            Write('Real poles at ',-p/2-dscr:10:3,-p/2+dscr:10:3);
}
        END;
    END;
END;


FUNCTION  LPCtoLog2Area(VAR a: predictor; ncoef: INTEGER): BOOLEAN;

    VAR
        i, j: INTEGER;
        r, rdiv, aj: DOUBLE;
        valid: BOOLEAN;

BEGIN
    valid := TRUE;
    a[0]  := 0;
    FOR i := ncoef DOWNTO 1 DO IF valid THEN BEGIN
        r := a[i];
        IF (r < -0.999) OR (r > 0.999) THEN
            valid := FALSE
        ELSE BEGIN
            rdiv := 1.0/(1.0 - r*r);
            FOR j := 1 TO i DIV 2 DO BEGIN
                aj     := rdiv*(a[j]   - r*a[i-j]);
                a[i-j] := rdiv*(a[i-j] - r*a[j]);
                a[j]   := aj;
            END;
            a[i] := Ln((1 + r)/(1 - r))/Ln(2);
        END;
    END;
    LPCtoLog2Area := valid;
END;


PROCEDURE Log2AreatoLPC(VAR a: predictor; ncoef: INTEGER);

    VAR
        i, j: INTEGER;
        r, aj: DOUBLE;

BEGIN
    a[0] := 1;
    FOR i := 1 TO ncoef DO BEGIN
        r := Exp(a[i]*Ln(2));
        r := (r - 1.0)/(r + 1.0);
        FOR j := 1 TO i DIV 2 DO BEGIN
            aj     := a[j]   + r*a[i-j];
            a[i-j] := a[i-j] + r*a[j];
            a[j]   := aj;
        END;
        a[i] := r;
    END;
END;

END.
